import asyncio
import datetime
import random
import websockets
import sys
import json
import signal
import sys

signal.signal(signal.SIGINT, signal.SIG_DFL)

RESET = '\033[0m'
GREEN = '\033[32m'

class WsServer:

    def __init__(self):
        print("I am listening at " + GREEN +  "ws://" + self.get_host() + ":" + self.get_port() + RESET + "...")
        self.loop = asyncio.get_event_loop()
       
    def get_port(self):
        return '5678'

    def get_host(self):
        return 'localhost'

    def init_server(self):
        return websockets.serve(self.handler, self.get_host(), self.get_port())

    def start(self):
        try:
            self.loop.run_until_complete(self.init_server())
            self.loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            pass

    async def handler(self, websocket, path):
        async for message in websocket:
            print('Path: ' + path + ', Received:'+ str(message) )
            request = json.loads(message)
            if   request["action"] == "message":
               obj = {}
               obj["aws"] = {}
               obj["aws"]["message"] = request["text"] + " Pong!"
               await websocket.send(json.dumps(obj)) 
            elif request["action"] == "listVpc":  
               response = self.listVpc(request)  
               await websocket.send(json.dumps(response)) 
            elif request["action"] == "close":
               print("Closing Socket...") 
               await websocket.close()
               print("Closed!") 
               sys.exit()

    def listVpc(self,request):
         response = {}
         response["aws"] = {}
         response["aws"]["vpcs"] = []
         response["aws"]["vpcs"].append({"vpcId":"vpc-123456789"})
         response["aws"]["vpcs"].append({"vpcId":"vpc-01457"})
         response["aws"]["vpcs"].append({"vpcId":"vpc-789456"})
         return response
               
            
if __name__ == '__main__':
  ws = WsServer()
  ws.start()
  




# coding=utf-8
#
# Author: Ualter Otoni Pereira
# ualter.junior@gmail.com
#
import os
import logging
import sys
from os.path import expanduser
import signal
from Functions import Functions
from Utils import Style, Utils
from Config import Config
from PyEc2Cp import PyEc2CP
from PyElbCp import PyElbCP
from PyAwsShell import PyAwsShell

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

signal.signal(signal.SIGINT, signal.SIG_DFL)

class PyAwsCp:
    def __init__(self):
        FILE_LOG = expanduser("~") + "/.pyawscp/" + "log"
        if not os.path.exists(FILE_LOG):
           os.makedirs(FILE_LOG)

        rootLogger = logging.getLogger()
        logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")

        fileHandler = logging.FileHandler("{0}/{1}.log".format(FILE_LOG, "py-awscp"))
        fileHandler.setFormatter(logFormatter)
        fileHandler.setLevel(logging.WARN)
        rootLogger.addHandler(fileHandler)

        #consoleHandler = logging.StreamHandler(sys.stdout)
        #consoleHandler.setFormatter(logFormatter)
        #consoleHandler.setLevel(logging.DEBUG)
        #rootLogger.addHandler(consoleHandler)

        self.start()

    def start(self):  
        if len(sys.argv) > 1:      
           if sys.argv[1] == "list" or sys.argv[1] == "help":
              Utils.clearScreen()
              Functions.showFunctions()
           elif sys.argv[1] == "-l" or sys.argv[1] == "-h":
              Utils.clearScreen()
              Functions.showFunctionsSummary()
           else:
              # One-way function execution
              s   = ""
              for arg in sys.argv:
                 s += arg
                 s += ","
              s = s[:len(s) - 1] 
              PyAwsShell().onewayFunction(s)  
           sys.exit()

        # Interactive 
        config = Config()
        config.interactive = True
        try:
           PyAwsShell().cmdloop()
        except KeyboardInterrupt:
           print("ERROR")
           pass   
    
if __name__ == '__main__':
    pyAwsCp = PyAwsCp()





#!/usr/bin/env python3

import socketserver
import http.server
import threading
import logging
import cgi
import signal
import sys, os, ast
import json
import configparser
from urllib.parse import unquote
from os.path import expanduser
from pygments import highlight, lexers, formatters

from pymxgraph.HttpServerServicesMock import HttpServerServicesMock
from pymxgraph.HttpServerServices import HttpServerServices

sys.path.append('..')
from Utils import Utils
from Emoticons import Emoticons

signal.signal(signal.SIGINT, signal.SIG_DFL)

PORT = 5679
LIST_VPC            = "listVpc"
LIST_SUBNET         = "listSubnet"
LIST_EC2            = "listEc2"
LIST_SECURITY_GROUP = "listSecurityGroup"

class HttpServerHandler(http.server.SimpleHTTPRequestHandler):

    VERBOSE     = False
    HTTP_SERVER = None

    def do_GET(self):
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        json_str = self.listVpc()
        self.wfile.write(json_str.encode(encoding='utf_8'))

    def do_OPTIONS(self):           
        self.send_response(200)
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
        self.send_header("Access-Control-Allow-Headers", "Content-Type, Authorization")
        self.end_headers()

    def do_POST(self):
        path        = self.path
        contentType = self.headers['Content-Type']

        if "/export" == path:
            self.exportImage(contentType)
        else:
            self.servicesCall(contentType)

    def exportImage(self, contentType):
        if "application/x-www-form-urlencoded" == contentType:
           contentLength = self.headers['Content-Length']
           request_bytes = self.rfile.read(int(contentLength))
           request_str   = request_bytes.decode('utf-8')
           request_str   = unquote(request_str)

           self.send_response(200)
           self.send_header('Access-Control-Allow-Origin', '*')
           self.send_header('Access-Control-Allow-Methods', 'POST')
           self.send_header("Content-Type", "image/png")
           self.end_headers()

           self.wfile.write(request_bytes)         

    def servicesCall(self, contentType):
        if "application/json" == contentType:
            contentLength = self.headers['Content-Length']
            request_bytes = self.rfile.read(int(contentLength))
            request_str   = request_bytes.decode('utf-8')
            request_json  = json.loads(request_str)

            if HttpServerHandler.VERBOSE:
               print(">>--->\033[93m REQUEST =================================================================\033[0m")
               print(highlight(Utils.dictToJson(request_json), lexers.JsonLexer(), formatters.TerminalFormatter()))
            
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Access-Control-Allow-Methods', 'POST')
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            response_json = self.processRequest(request_json)

            if HttpServerHandler.VERBOSE:
               print(">>--->\033[93m RESPONSE =================================================================\033[0m")
               print(highlight(response_json, lexers.JsonLexer(), formatters.TerminalFormatter()))
            
            self.wfile.write(response_json.encode(encoding='utf_8'))         
        else:
            self.send_response(400)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Access-Control-Allow-Methods', 'POST')
            self.send_header("Content-Type", "application/json")
            self.end_headers()

    def showAction(self,action):
        if HttpServerHandler.VERBOSE:
           print(">>--->\033[93m ACTION   =========   \033[32m{}\033[93m".format(action))

    def processRequest(self, request_json):
        response_json = ""
        if  LIST_VPC == request_json["action"]:
            self.showAction("List VPCs")
            response_json = self.listVpc()
        elif LIST_SUBNET == request_json["action"]:
            self.showAction("List Subnets")
            response_json = self.listSubnet(request_json)
        elif LIST_EC2 == request_json["action"]:
            self.showAction("List EC2s")
            response_json = self.listEc2(request_json)
        elif LIST_SECURITY_GROUP == request_json["action"]:
            self.showAction("List Security Groups")
            response_json = self.listSecurityGroup(request_json)    
        else:
            response = {}
            response["aws"] = {}
            response["aws"]["status"] = "Action " + request_json["action"] + " not recognized!"
            response_json = json.dumps(response)
        return response_json      
                

    def listVpc(self):
        httpServices  = HttpServerServices()
        #httpServicesMock = HttpServerServicesMock()
        return httpServices.listVpc()

    def listEc2(self, request_json):
        httpServices = HttpServerServices()
        return httpServices.listEc2(request_json)

    def listSubnet(self, request_json):
        httpServices = HttpServerServices()
        #httpServicesMock = HttpServerServicesMock()
        return httpServices.listSubnet(request_json)

    def listSecurityGroup(self, request_json):
        httpServices = HttpServerServices()
        return httpServices.listSecurityGroup(request_json)    

class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass

shutdown_evt = threading.Event()

class HttpServer:
    def __init__(self, verbose):
        self.httpd   = None
        self.verbose = verbose
        HttpServerHandler.VERBOSE = self.verbose
    def run(self):
        self.httpd = ThreadedTCPServer(('127.0.0.1', PORT), HttpServerHandler)
        server_thread = threading.Thread(target=self.httpd.serve_forever)
        server_thread.daemon = True
        server_thread.start()
        print(Emoticons.ok() + " \033[93mAWS Navigator serving\033[0m")
        print(Emoticons.ok() + " Press \033[93mENTER\033[0m to shutdown...")
        input("")
        shutdown_evt.set()
        self.httpd.shutdown()
        self.httpd.server_close()

if __name__ == '__main__':
   httpServer = HttpServer(True)
   httpServer.run()


#!/usr/bin/env python3

import boto3
import logging
import sys
import json
from botocore.exceptions import ClientError
from datetime import datetime
# Run temporarily local (python command)
#from pyawscp.utils import utils
from Functions import Functions
from Utils import Utils, Style
from PrettyTable import PrettyTable
from Emoticons import Emoticons
# For test local
from Config import Config
from TableArgs import TableArgs
from pygments import highlight, lexers, formatters

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)


class PyEc2CP:

    config = None

    def __init__(self, config):
        self.config = config

    # A main function called from user
    # Check if a specific subnetId is Public looking up its Route Table for a IGW-* 0.0.0.0/0
    # arguments:
    #      subnetId: mandatory
    def subnetIsPublic(self):
        ec2      = self.botoSession().client('ec2')
        subnetId = self.config.commandArguments

        if not subnetId:
           resultTxt = "Where is the SubnetId? You didn't tell me which one " + Emoticons.ops() +  Style.RESET
           return Utils.formatResult(Functions.SUBNET_IS_PUBLIC,resultTxt, self.config, "", True)

        filters=[{'Name': 'association.subnet-id','Values': [subnetId]}, 
                 {'Name': 'route.destination-cidr-block','Values': ['0.0.0.0/0']},
                 {'Name': 'route.gateway-id','Values': ['igw-*']}]
        if self.config.awsTagsToFilter():
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               ) 
        route_tables = ec2.describe_route_tables(Filters=filters)

        jsonResult = ""
        if self.config.printResults: 
           jsonResult = Utils.dictToJson(route_tables)

        routeTableIgw = route_tables['RouteTables']
        if len(routeTableIgw) >= 1:
           resultTxt = "Subnet " + Style.GREEN + subnetId + Style.RESET + " is " + Style.GREEN + "PUBLIC" +  Style.RESET + "\n"

           printSeparator = False
           for route in routeTableIgw[0]["Routes"]:
               if route["DestinationCidrBlock"] == '0.0.0.0/0':
                  if route["State"]:
                     if not printSeparator:
                        resultTxt = resultTxt + Utils.separator()
                        printSeparator = True
                     resultTxt = resultTxt + "\n"
                     resultTxt = resultTxt + ' '.ljust(2,' ') + "- State...........: " + Style.GREEN + route['State'] + Style.RESET 
                  if route["GatewayId"]:
                     if not printSeparator:
                        resultTxt = resultTxt + Utils.separator()
                        printSeparator = True
                     resultTxt = resultTxt + "\n"
                     resultTxt = resultTxt + ' '.ljust(2,' ') + "- GatewayId.......: " + Style.GREEN + route['GatewayId'] + Style.RESET    
                  if routeTableIgw[0]["VpcId"]:
                     if not printSeparator:
                        resultTxt = resultTxt + Utils.separator()
                        printSeparator = True
                     resultTxt = resultTxt + "\n"
                     resultTxt = resultTxt + ' '.ljust(2,' ') + "- Vpc Id..........: " + Style.GREEN + routeTableIgw[0]['VpcId'] + Style.RESET   
           return Utils.formatResult(Functions.FUNCTIONS[Functions.SUBNET_IS_PUBLIC]["name"],resultTxt, self.config, jsonResult, True, None)
        else:
           resultTxt = ""
           subnet = ec2.describe_subnets(Filters=[{'Name':'subnet-id','Values':[subnetId]}])
           if len(subnet['Subnets']) == 0:
               resultTxt = "Subnet " + Style.GREEN + subnetId + Style.RED + " not found! " + Emoticons.ops() +  Style.RESET
           else:   
               if self.config.awsTagsToFilter():
                  resultTxt = "Subnet " + Style.GREEN + subnetId + Style.RESET + " is NOT " + Style.GREEN + "PUBLIC" +  Style.RESET + ", or either did not pass the Filters Tag! " + + Emoticons.thumbsUp()
               else:   
                  resultTxt = "Subnet " + Style.GREEN + subnetId + Style.RESET + " is NOT " + Style.GREEN + "PUBLIC" +  Style.RESET + " " + Emoticons.thumbsUp() 
           return Utils.formatResult(Functions.FUNCTIONS[Functions.SUBNET_IS_PUBLIC]["name"],resultTxt, self.config, "", True, None)     
    
    # Reusable Function, used by: 
    # -lisSubtnetsVpc
    def _checkSubnetIsPublic(self, subnetId, extraFilters):
        ec2 = self.botoSession().client('ec2')
        filters=[{'Name': 'association.subnet-id','Values': [subnetId]}, 
                 {'Name': 'route.destination-cidr-block','Values': ['0.0.0.0/0']},
                 {'Name': 'route.gateway-id','Values': ['igw-*']}
        ]
        route_tables = ec2.describe_route_tables(Filters=filters)
        routeTableIgw = route_tables['RouteTables']
        if len(routeTableIgw) >= 1:
            return True
        else:
            return False    

    # List all Subnets from a VPC
    def listSubnetsVpc(self):
        ec2   = self.botoSession().client('ec2')

        vpcId     = ""
        isPublic  = False
        tableArgs = TableArgs()
        tableArgs.setArguments(self.config.commandArguments) 

        if "," in self.config.commandArguments:
           vpcId = self.config.commandArguments.split(",")[0]
           tableArgs.setArguments(self.config.commandArguments)
           if "ispublic" in self.config.commandArguments:
               isPublic = True
        else:
           vpcId = tableArgs.cleanPipelineArguments()

        if not vpcId:
           resultTxt = "Where is the VpcId? You didn't tell me which VPC " + Emoticons.ops() +  Style.RESET
           return Utils.formatResult(Functions.LIST_SUBNETS_VPC,resultTxt, self.config, "", True, tableArgs)

        filters=[
            {
                'Name': 'vpc-id',
                'Values': [
                    vpcId
                ]
            }
        ]
        if self.config.awsTagsToFilter():
           # Environment Tags set
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               )

        # Tags from command line arguments
        if len(tableArgs.tagsTemp) > 0:
           for tag in tableArgs.tagsTemp: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]}
               )

        subnets = ec2.describe_subnets( Filters=filters )

        jsonResult = ""
        if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile: 
           jsonResult = Utils.dictToJson(subnets)

        resultTxt = ""
        if len(subnets["Subnets"]) >= 1:
            resultTxt = "The VPC " + Style.GREEN + vpcId + Style.RESET + " has " + Style.GREEN + str(len(subnets["Subnets"])) + Style.RESET + " Subnets\n"
            resultTxt = resultTxt + Utils.separator()  + "\n\n"

            idx_lin=0
            header      = ["#","Subnet Id", "Availability Zone", "CIDR Block", "Available IP Address Count"]
            if isPublic:
                header.append("Is Public?")
            if tableArgs.showTags:
                header.append("Tags")    
            prettyTable = PrettyTable(header)
            
            for subnet in subnets["Subnets"]:
                idx_lin += 1

                columns = [ str(idx_lin), subnet['SubnetId'], subnet['AvailabilityZone'],  subnet['CidrBlock'], subnet['AvailableIpAddressCount'] ]

                if isPublic:
                   resultIsPublic = self._checkSubnetIsPublic(subnet['SubnetId'], None)
                   result = "--- "
                   if resultIsPublic:
                      result = "PUBLIC"
                   columns.append(result)
                if tableArgs.showTags:
                   tags = Utils.formatPrintTags(subnet['Tags'])
                   columns.append(tags)

                prettyTable.addRow(columns)

            if (int(tableArgs.sortCol) - 1) > len(columns):
                tableArgs.sortCol = "1"
            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
            prettyTable.ascendingOrder(not tableArgs.desc)
            result = "VPC...: " + Style.GREEN + vpcId + Style.RESET + "\n\n" + prettyTable.printMe(self.config.tableLineSeparator, tableArgs)
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_SUBNETS_VPC]["name"], result, self.config, jsonResult, True, tableArgs)
        else:
           resultTxt = ""
           try: 
             vpc = ec2.describe_vpcs(VpcIds=[vpcId])
             if not self.config.awsTagsToFilter():
                resultTxt = "VPC was found, but there's no Subnet associated. " + Emoticons.ops()
             else:   
                resultTxt = "VPC was found, but there's no Subnet associated or either did not pass the Filters Tag! " + Emoticons.ops()
           except ClientError as e:
             if "InvalidVpcID.NotFound" == e.response['Error']['Code']:
                resultTxt = "VPC with the Id " + Style.GREEN + vpcId + Style.RESET + " was not found! " + Emoticons.ops() +  Style.RESET
             else:    
                resultTxt = "Exception! ❌💣 " + Style.GREEN + e.response['Error']['Code'] + Style.RESET
           except:
             resultTxt = sys.exc_info + " ❌💣 "
           return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_SUBNETS_VPC]["name"], resultTxt, self.config, "", True, tableArgs)

    # List all VPCs
    def listVpc(self):
        ec2 = self.botoSession().client('ec2')

        tableArgs = TableArgs()
        tableArgs.setArguments(self.config.commandArguments)

        listSubnets    = False
        subnetIsPublic = False

        if "subnets" in self.config.commandArguments:
           listSubnets    = True
        if "ispublic" in self.config.commandArguments:
           subnetIsPublic = True   

        filters = []
        if self.config.awsTagsToFilter():
           # Tags from Environment
           for tag in self.config.awsTags: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]}
               )

        # Tags from command line arguments
        if len(tableArgs.tagsTemp) > 0:
           for tag in tableArgs.tagsTemp: 
               filters.append(
                   {'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]}
               )
        vpcs = ec2.describe_vpcs(Filters=filters)

        jsonResult = ""
        if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile:
           jsonResult = Utils.dictToJson(vpcs)

        resultTxt = ""
        if len(vpcs["Vpcs"]) >= 1:
            resultTxt = "List of VPCs, total of " + Style.GREEN + str(len(vpcs["Vpcs"])) + Style.RESET + "\n"
            resultTxt = resultTxt + Utils.separator()  + "\n\n"

            idx_lin=0
            header = ["#","VPC Id", "CIDR Block", "Default"]
            if listSubnets:
               header = ["#","VPC Id", "CIDR Block", "Default", "Subnet Id", "Subnet CIDR Block", "AZ", "Available IP Count"] 
               if subnetIsPublic:
                  header.append("Is Public?")
            if tableArgs.showTags:
               header.append("VPC Tags")       

            prettyTable = PrettyTable(header)
            for vpc in vpcs["Vpcs"]:
                idx_lin += 1

                vpcId     = vpc['VpcId']
                cidrBlock = vpc['CidrBlock']
                default   = "Default" if vpc['IsDefault'] else "---"
                columns   = None
                # Asked to list the subnets of each VPC?
                if listSubnets:
                   subnets = ec2.describe_subnets( Filters=[{'Name': 'vpc-id','Values': [vpc['VpcId']]}])
                   idx_subnet_lin = 0

                   for subnet in subnets["Subnets"]:
                       idx_subnet_lin += 1
                       lblSubnet = subnet['SubnetId'] + "  (" + str(idx_subnet_lin) + ")"
                       cidrBlockSubnet = subnet['CidrBlock']
                       az = subnet['AvailabilityZone']
                       availableIpdAddressCount = subnet['AvailableIpAddressCount']

                       columns = [ idx_lin, vpcId, cidrBlock, default, lblSubnet, cidrBlockSubnet, az, availableIpdAddressCount ]

                       # Asked to inform if the Subnet is Public? 
                       if subnetIsPublic:
                          resultSubnetIsPublic = "--- "
                          if self._checkSubnetIsPublic(subnet['SubnetId'], None) :
                             resultSubnetIsPublic = "PUBLIC"
                          columns.append(resultSubnetIsPublic)
                       if tableArgs.showTags:
                          tags = Utils.formatPrintTags(vpc['Tags'])
                          columns.append(tags)   

                       prettyTable.addRow(columns)

                   prettyTable.addSeparatorGroup()    
                else:   
                   columns = [ idx_lin, vpcId, cidrBlock, default ]
                   if tableArgs.showTags:
                      tags = Utils.formatPrintTags(vpc['Tags'])
                      columns.append(tags)
                   prettyTable.addRow(columns)

            if (int(tableArgs.sortCol) - 1) > len(columns):
                sortCol = "1"

            prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
            prettyTable.ascendingOrder(not tableArgs.desc)
            result = "VPCs " + "\n\n" + prettyTable.printMe(self.config.tableLineSeparator, tableArgs)
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_VPC]["name"], result, self.config, jsonResult, True, tableArgs)    
        else:
            return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_VPC]["name"], " No VPC found " + Emoticons.ops(), self.config, jsonResult, True, tableArgs)         

    # List all Ec2 Instances
    def listEc2(self):
       ec2              = self.botoSession().client('ec2')
       instanceId       = None
       privateIPFilter  = None
       vpcIdFilter      = None
       subnetIdFilter   = None
       showSg           = False
       showVpc          = False
       showSubnet       = False
       showImage        = False
       showCpu          = False
       hideName         = False

       tableArgs = TableArgs()
       tableArgs.setArguments(self.config.commandArguments)
       if "," in self.config.commandArguments:
          for arg in self.config.commandArguments.split(","):
              if "i-" in arg:
                 instanceId = arg
              elif arg == "sg":
                 showSg = True
              elif arg == "vpc":
                 showVpc = True    
              elif arg == "subnet":
                 showSubnet = True   
              elif arg == "image":
                 showImage = True 
              elif arg == "cpu":
                 showCpu = True
              elif arg == "noname":
                 hideName = True
              elif arg.startswith("vpc-"):
                 vpcIdFilter = arg
              elif arg.startswith("subnet-"):
                 subnetIdFilter = arg              
              else:
                 # Check IP
                 if Utils.isNumber(arg[0:2]):
                    privateIPFilter = arg

       else:
          arg = self.config.commandArguments
          if "i-" in arg:
            instanceId = arg
          elif arg == "sg":
             showSg = True
          elif arg == "vpc":
             showVpc = True    
          elif arg == "subnet":
             showSubnet = True   
          elif arg == "image":
             showImage = True 
          elif arg == "cpu":
             showCpu = True
          elif arg == "noname": 
             hideName = True
          elif arg.startswith("vpc-"):
             vpcIdFilter = arg
             print(arg)
          elif arg.startswith("subnet-"):
             subnetIdFilter = arg   
          else:
             # Check IP
             if Utils.isNumber(arg[0:2]):
                privateIPFilter = arg


       filters=[]   

       if instanceId:
          filters.append({'Name': 'instance-id','Values': [instanceId]})   
       if privateIPFilter:   
          filters.append({'Name': 'private-ip-address','Values': [privateIPFilter]})
       if vpcIdFilter:   
          filters.append({'Name': 'vpc-id','Values': [vpcIdFilter]})
       if subnetIdFilter:   
          filters.append({'Name': 'subnet-id','Values': [subnetIdFilter]})         

       if self.config.awsTagsToFilter():
           # Environment Tags set
           for tag in self.config.awsTags: 
               filters.append({'Name':'tag:' + tag,'Values':[self.config.awsTags[tag]]})   

       # Tags from command line arguments
       if len(tableArgs.tagsTemp) > 0:
          for tag in tableArgs.tagsTemp: 
              filters.append({'Name':'tag:' + tag,'Values':[tableArgs.tagsTemp[tag]]})        

       reservations = ec2.describe_instances( Filters=filters )        

       jsonResult = ""
       if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile: 
          jsonResult = Utils.dictToJson(reservations)

       if reservations and len(reservations["Reservations"]) > 0:
  
          idx_lin = 0

          if hideName:
             header = ["#","Instance Id",       "Type","Launch Time", "Private IP", "Public IP", "State"]
          else:   
             header = ["#","Instance Id","Name","Type","Launch Time", "Private IP", "Public IP", "State"]

          if showSg:             
             header.append("Security Groups")
          if showImage:          
             header.append("Image Id")
          if showCpu:            
             header.append("CPU Options")
          if showVpc:            
             header.append("VPC")
          if showSubnet:         
             header.append("Subnet")
          if tableArgs.showTags: 
             header.append("Tags")

          prettyTable = PrettyTable(header)

          totalListed = 0

          for instanceGroup in reservations["Reservations"]:
              for instance in instanceGroup["Instances"]:
                  idx_lin     += 1
                  totalListed += 1

                  if not hideName:
                     name = ""
                     for t in instance['Tags']:
                         if t["Key"] == "Name":
                            name = t["Value"]

                  launchTimeDate = ""
                  if instance['LaunchTime']:
                     launchTimeDate = instance['LaunchTime'].strftime("%d-%b-%y %H:%M")

                  privateIp = instance['PrivateIpAddress']
                  publicIp  = instance["PublicIpAddress"] if "PublicIpAddress" in instance else ""
                  state     = instance["State"]["Name"]
                  state     = Style.CYAN + state + Style.RESET if state == "stopped" else state

                  if hideName:
                     columns = [ str(idx_lin), str(instance['InstanceId']), instance['InstanceType'], launchTimeDate, privateIp, publicIp, state ]
                  else:   
                     columns = [ str(idx_lin), str(instance['InstanceId']), name, instance['InstanceType'], launchTimeDate, privateIp, publicIp, state ]

                  if showSg:
                     securityGroups = ""
                     for sg in instance["SecurityGroups"]:             
                         if securityGroups != "": securityGroups += ","
                         securityGroups += sg["GroupId"]
                     columns.append(securityGroups)    
                  if showImage:          
                     columns.append(instance["ImageId"])       
                  if showCpu:   
                     cpu  = Style.MAGENTA + "Core:"    + Style.GREEN + str(instance["CpuOptions"]["CoreCount"]) + Style.RESET + " "
                     cpu += Style.MAGENTA + "Threads:" + Style.GREEN + str(instance["CpuOptions"]["ThreadsPerCore"])
                     columns.append(cpu)       
                  if showVpc:            
                     vpc = instance["VpcId"]
                     columns.append(vpc)
                  if showSubnet:         
                     subnet = instance["SubnetId"]
                     columns.append(subnet)
                  if tableArgs.showTags:
                     tags = Utils.formatPrintTags(instance['Tags'])
                     columns.append(tags)

                  prettyTable.addRow(columns)

          resultTxt = ""

          if instanceId:
             if resultTxt != "": resultTxt = resultTxt + "\n"
             resultTxt = resultTxt + "      Instance Id........: " + Style.GREEN + instanceId + Style.RESET
          if privateIPFilter:   
             if resultTxt != "": resultTxt = resultTxt + "\n"
             resultTxt = resultTxt + "      Private IP Address.: " + Style.GREEN + privateIPFilter + Style.RESET
          if vpcIdFilter:   
             if resultTxt != "": resultTxt = resultTxt + "\n"
             resultTxt = resultTxt + "      VPC Id.............: " + Style.GREEN + vpcIdFilter + Style.RESET
          if subnetIdFilter:   
             if resultTxt != "": resultTxt = resultTxt + "\n"
             resultTxt = resultTxt + "      Subnet Id..........: " + Style.GREEN + subnetIdFilter + Style.RESET

          resultTxt = resultTxt + "\n\n Total of EC2s...: " + Style.GREEN + format(totalListed,",") + Style.RESET   

          if (int(tableArgs.sortCol) - 1) > len(columns):
              tableArgs.sortCol = "1"
          prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
          prettyTable.ascendingOrder(not tableArgs.desc)
          result = resultTxt + "\n\n" + prettyTable.printMe(self.config.tableLineSeparator, tableArgs)
          return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_EC2]["name"], result, self.config,jsonResult, True, tableArgs)    
       else:
          resultTxt = ""
          resultTxt = "No EC2 Instance was found. " + Emoticons.ops()
          return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_EC2]["name"], resultTxt, self.config, "", True, tableArgs)

    def botoSession(self):
        return self.config.botoSession()

if __name__ == '__main__':
    config = Config()
    config.awsProfile = "ecomm"
    config.awsRegion = "eu-central-1"
    config.printResults = True
    #config.awsTags["Environment"] = "production"
    config.tableLineSeparator = False
    
    
    # subnets, is-public (only with subnets active)
    # sort1,desc (when requested together with subnets, the group lines will be omitted)
    if len(sys.argv) > 1:
       config.commandArguments = sys.argv[1]
    pyEc2CP = PyEc2CP(config)
    report = pyEc2CP.listEc2()
    Utils.addToClipboard(report)
    print(report)

    # TODO:
    # findEc2 (by IP, by InstanceId, by Tag, etc.)
    # Result show: IP, t3.medium, region, subnet, vpc, security group, (attached Load Balancer optional), (Security Groups Rules optional)



#!/usr/bin/env python3

import socketserver
import http.server
import logging
import cgi
import signal
import sys, os, ast
import json
import configparser
import datetime
from os.path import expanduser

PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))
from Config import Config
from Utils import Utils


LIST_VPC            = "listVpc"
LIST_SUBNET         = "listSubnet"
LIST_EC2            = "listEc2"
LIST_SECURITY_GROUP = "listSecurityGroup"

class HttpServerServices:

    def __init__(self):
        FILE_INI = expanduser("~") + "/.pyawscp/" + "pyawscp.ini"
        self.config = Config()
        if os.path.exists(FILE_INI):
           configFileIni = configparser.ConfigParser()
           configFileIni.read(FILE_INI)
           self.config.awsProfile         = configFileIni["PREFERENCES"]["aws-profile"]
           self.config.awsRegion          = configFileIni["PREFERENCES"]["aws-region"]
           self.config.awsTags            = ast.literal_eval(configFileIni["PREFERENCES"]["aws-tags"])
           self.config.printResults       = configFileIni["PREFERENCES"]["print-results"] in ['True','true']
           self.config.tableLineSeparator = configFileIni["PREFERENCES"]["table-line-separator"] in ['True','true']

    def listVpc(self):
        ec2Client              = self.botoSession().client('ec2')
        vpcs                   = ec2Client.describe_vpcs()
        response               = self.buildAwsResponse(LIST_VPC)
        response["aws"]["vpc"] = []
        for vpc in vpcs["Vpcs"]:

            tags = []
            for tag in vpc["Tags"]:
                tags.append({"key":tag["Key"],"value":tag["Value"]})

            response["aws"]["vpc"].append(
                {
                    "vpcId":vpc["VpcId"],
                    "cidrBlock":vpc["CidrBlock"],
                    "tags": tags
                }
            )
        json_str = json.dumps(response)
        return json_str
    
    def listSubnet(self, request_json):
        ec2Client = self.botoSession().client('ec2')
        params    = self.buildParams(request_json)
        filters=[
            {
                'Name': 'vpc-id',
                'Values': [
                    params["vpcId"]
                ]
            }
        ]
        subnets  = ec2Client.describe_subnets( Filters=filters )
        response = self.buildAwsResponse(LIST_SUBNET)
        response["aws"]["vpc"]            = {}
        response["aws"]["vpc"]["vpcId"]   = params["vpcId"]
        response["aws"]["vpc"]["subnets"] = []

        for subnet in subnets["Subnets"]:
            tags = []
            for tag in subnet["Tags"]:
                tags.append({"key":tag["Key"],"value":tag["Value"]})

            isPublic = self._checkSubnetIsPublic(ec2Client, subnet["SubnetId"])    
            public   = "public" if isPublic else "private"

            subnetJson = {
                "subnetId":subnet["SubnetId"],
                "vpcId":subnet["VpcId"],
                "scope":public,
                "availabilitZone":subnet["AvailabilityZone"],
                "availabilitZoneId":subnet["AvailabilityZoneId"] if "AvailabilityZoneId" in subnet else "",
                "availableIpAddressCount":subnet["AvailableIpAddressCount"],
                "cidrBlock":subnet["CidrBlock"],
                "defaultForAz":subnet["DefaultForAz"],
                "state":subnet["State"],
                "subnetArn":subnet["SubnetArn"] if "SubnetArn" in subnet else "",
                "tags": tags
            }
            
            response["aws"]["vpc"]["subnets"].append(subnetJson)
        json_str = json.dumps(response)
        return json_str

    def _checkSubnetIsPublic(self, ec2Client, subnetId):
        filters=[{'Name': 'association.subnet-id','Values': [subnetId]}, 
                 {'Name': 'route.destination-cidr-block','Values': ['0.0.0.0/0']},
                 {'Name': 'route.gateway-id','Values': ['igw-*']}
        ]
        route_tables = ec2Client.describe_route_tables(Filters=filters)
        routeTableIgw = route_tables['RouteTables']
        if len(routeTableIgw) >= 1:
            return True
        else:
            return False

    def listEc2(self, request_json):
        ec2Client = self.botoSession().client('ec2')
        params    = self.buildParams(request_json)
        filters=[
            {
                'Name': 'subnet-id',
                'Values': [
                    params["subnetId"]
                ]
            }
        ]
        ec2s  = ec2Client.describe_instances( Filters=filters )
        response = self.buildAwsResponse(LIST_EC2)
        response["aws"]["ec2"] = {}
        response["aws"]["ec2"]["vpcId"]     = params["vpcId"]
        response["aws"]["ec2"]["subnetId"]  = params["subnetId"]
        response["aws"]["ec2"]["instances"] = []

        for reservations in ec2s["Reservations"]:
           for instance in reservations["Instances"]:
               tags = []
               for tag in instance["Tags"]:
                   tags.append({"key":tag["Key"],"value":tag["Value"]})
   
               securityGroups = []
               for sg in instance["SecurityGroups"]:
                   securityGroups.append({"groupId":sg["GroupId"],"groupName":sg["GroupName"]})
   
               blockDeviceMappings = []
               for bdm in instance["BlockDeviceMappings"]:
                   blockDeviceMappings.append({
                                "deviceName":bdm["DeviceName"],
                                "ebs":{
                                    "attachTime":bdm["Ebs"]["AttachTime"],
                                    "deleteOnTermination":bdm["Ebs"]["DeleteOnTermination"],
                                    "status":bdm["Ebs"]["Status"],
                                    "volumeId":bdm["Ebs"]["VolumeId"],
                                }
                              })        
   
               publicIpAddress = ""
               if "PublicIpAddress" in instance:
                   publicIpAddress = instance["PublicIpAddress"]

               cpuValue = {}
               if "CpuOptions" in instance:
                   cpuValue["coreCount"]      = instance["CpuOptions"]["CoreCount"]
                   cpuValue["threadsPerCore"] = instance["CpuOptions"]["ThreadsPerCore"]

               instanceJson = {
                   "subnetId":instance["SubnetId"],
                   "vpcId":instance["VpcId"],
                   "instanceId":instance["InstanceId"],
                   "imageId":instance["ImageId"],
                   "instanceType":instance["InstanceType"],
                   "launchTime":instance["LaunchTime"],
                   "privateDnsName":instance["PrivateDnsName"],
                   "privateIpAddress":instance["PrivateIpAddress"],
                   "publicIpAddress":publicIpAddress,
                   "publicDnsName":instance["PublicDnsName"],
                   "state":{"code":instance["State"]["Code"],"name":instance["State"]["Name"]},
                   "ebsOptimized":instance["EbsOptimized"],
                   "rootDeviceName":instance["RootDeviceName"],
                   "rootDeviceType":instance["RootDeviceType"],
                   "securityGroups":securityGroups,
                   "cpu":cpuValue,
                   "blockDeviceMappings":blockDeviceMappings,
                   "tags": tags
               }
               
               response["aws"]["ec2"]["instances"].append(instanceJson)

        json_str = json.dumps(response, default = self.myConverter)
        return json_str

    def listSecurityGroup(self, request_json):
        ec2Client = self.botoSession().client('ec2')
        params    = self.buildParams(request_json)
        filters=[
            {
                'Name': 'group-id',
                'Values': [
                    params["groupId"]
                ]
            }
        ]
        securityGroups  = ec2Client.describe_security_groups( Filters=filters )
        response = self.buildAwsResponse(LIST_SECURITY_GROUP)
        response["aws"]["ec2"] = {}
        response["aws"]["ec2"]["securityGroup"] = []

        for securityGroup in securityGroups["SecurityGroups"]:

            tags = []
            for tag in securityGroup["Tags"]:
                tags.append({"key":tag["Key"],"value":tag["Value"]})


            ingresses = []
            for ingress in securityGroup["IpPermissions"]:
                ipRanges = []
                for ipRange in ingress["IpRanges"]:
                    ipRanges.append({
                        "cidrIp":ipRange["CidrIp"],
                        "description":ipRange["Description"],
                    })

                ingresses.append({
                    "fromPort":ingress["FromPort"],
                    "ipProtocol":ingress["IpProtocol"],
                    "ipRanges": ipRanges,
                    "toPort":ingress["ToPort"]
                })

            egresses = []
            for egress in securityGroup["IpPermissionsEgress"]:
                ipRanges = []
                for ipRange in egress["IpRanges"]:
                    ipRanges.append({
                        "cidrIp":ipRange["CidrIp"],
                        "description":ipRange["Description"],
                    })

                egresses.append({
                    "ipProtocol":egress["IpProtocol"],
                    "ipRanges": ipRanges
                })        


            sgJson = {
                   "groupId":securityGroup["GroupId"],
                   "description":securityGroup["Description"],
                   "vpcId":securityGroup["VpcId"],
                   "ingress": ingresses,
                   "egress": egresses,
                   "tags": tags
            }
               
            response["aws"]["ec2"]["securityGroup"].append(sgJson) 

        json_str = json.dumps(response, default = self.myConverter)
        return json_str          

        """ 
        for reservations in ec2s["Reservations"]:
           for instance in reservations["Instances"]:
               tags = []
               for tag in instance["Tags"]:
                   tags.append({"key":tag["Key"],"value":tag["Value"]})
   
               securityGroups = []
               for sg in instance["SecurityGroups"]:
                   securityGroups.append({"groupId":sg["GroupId"],"groupName":sg["GroupName"]})
   
               blockDeviceMappings = []
               for bdm in instance["BlockDeviceMappings"]:
                   blockDeviceMappings.append({
                                "deviceName":bdm["DeviceName"],
                                "ebs":{
                                    "attachTime":bdm["Ebs"]["AttachTime"],
                                    "deleteOnTermination":bdm["Ebs"]["DeleteOnTermination"],
                                    "status":bdm["Ebs"]["Status"],
                                    "volumeId":bdm["Ebs"]["VolumeId"],
                                }
                              })        
   
               publicIpAddress = ""
               if "PublicIpAddress" in instance:
                   publicIpAddress = instance["PublicIpAddress"]

               instanceJson = {
                   "subnetId":instance["SubnetId"],
                   "vpcId":instance["VpcId"],
                   "instanceId":instance["InstanceId"],
                   "imageId":instance["ImageId"],
                   "instanceType":instance["InstanceType"],
                   "launchTime":instance["LaunchTime"],
                   "privateDnsName":instance["PrivateDnsName"],
                   "privateIpAddress":instance["PrivateIpAddress"],
                   "publicIpAddress":publicIpAddress,
                   "publicDnsName":instance["PublicDnsName"],
                   "state":{"code":instance["State"]["Code"],"name":instance["State"]["Name"]},
                   "ebsOptimized":instance["EbsOptimized"],
                   "rootDeviceName":instance["RootDeviceName"],
                   "rootDeviceType":instance["RootDeviceType"],
                   "securityGroups":securityGroups,
                   "cpu":{"coreCount":instance["CpuOptions"]["CoreCount"],"threadsPerCore":instance["CpuOptions"]["ThreadsPerCore"]},
                   "blockDeviceMappings":blockDeviceMappings,
                   "tags": tags
               }
               
               response["aws"]["ec2"]["instances"].append(instanceJson)

        json_str = json.dumps(response, default = self.myConverter)
        return json_str  """  

    def myConverter(self,o):
        if isinstance(o, datetime.datetime):
           #return o.__str__()    
           return o.strftime('%d-%m-%Y %H:%M:%S')

    def botoSession(self):
        return self.config.botoSession()

    def buildAwsResponse(self, action):
        response = {}
        response["aws"] = {}
        response["aws"]["status"] = "OK"
        response["aws"]["action"] = action
        return response

    def buildParams(self, json):
        params = {}
        for param in json["parameters"]:
           params[param["name"]] = {}
           params[param["name"]] = param["value"]
        return params

    def getParamValue(self, request_json, paraName): 
        for params in request_json["parameters"]:
            if params["name"] == paraName:
               return params["value"]
        return None    

if __name__ == '__main__':
   httpServices = HttpServerServices()
   #print(httpServices.listVpc())
   #print(httpServices.listSubnet({'action': 'listSubnet', 'parameters': [{'name': 'vpcId', 'value': 'vpc-090a0f7286ab4281c'}]}))
   #print(httpServices.listEc2({'action': 'listEc2', 'parameters': [{'name': 'vpcId', 'value': 'vpc-01c6810fb391d215b'}, {'name': 'subnetId', 'value': 'subnet-0313b32f6f116c352'}]}))
   print(httpServices.listSecurityGroup({'action': 'listSecurityGroup', 'parameters': [{'name': 'groupId', 'value': 'sg-07dcfaec3ee966a3d'}]}))

      



import boto3
import logging
import sys
import json
from botocore.exceptions import ClientError
from datetime import datetime
from arnparse import arnparse
# Run temporarily local (python command)
#from pyawscp.utils import utils
from Functions import Functions
from TableArgs import TableArgs
from Utils import Utils, Style
from PrettyTable import PrettyTable
from Emoticons import Emoticons
from Config import Config
from pygments import highlight, lexers, formatters

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)


class PyElbCP:

    config = None

    def __init__(self, config):
        self.config = config

    # A main function called from user
    # List the Target Groups, Health Checks and its connected ELBs
    def listTargetGroups(self):
        elbv2       = self.botoSession().client('elbv2')
        tgArn       = []
        elbArn      = None
        showElbArn  = False
        showTgArn   = False
        healthInfo  = False

        tableArgs = TableArgs()

        tableArgs.setArguments(self.config.commandArguments)
        if "," in self.config.commandArguments: 
           for arg in self.config.commandArguments.split(","):
               if "health" in arg:
                  healthInfo = True
               elif "elbarn" in arg:
                  showElbArn = True
               elif "tgarn" in arg:
                  showTgArn = True      
               elif ":targetgroup/" in arg:
                  if ";" in arg: 
                     for tg in arg.split(";"): 
                        tgArn.append(tg)
                  else:   
                     tgArn.append(arg)   
               elif ":loadbalancer/" in arg:
                  elbArn = arg
        else:
           if "health" in self.config.commandArguments:
               healthInfo = True
           if ":targetgroup/" in self.config.commandArguments:
               if ";" in self.config.commandArguments: 
                  for tg in self.config.commandArguments.split(";"): 
                        tgArn.append(tg)
               else:   
                  tgArn.append(self.config.commandArguments)    
           if ":loadbalancer/" in self.config.commandArguments:
               elbArn = self.config.commandArguments      
           if "elbarn" in self.config.commandArguments:
               showElbArn = True
           if "tgarn" in self.config.commandArguments:
               showTgArn = True          

        target_groups = None
        if tgArn:
           target_groups = elbv2.describe_target_groups(TargetGroupArns=tgArn)
        elif elbArn:
           target_groups = elbv2.describe_target_groups(LoadBalancerArn=elbArn)
        else:
           target_groups = elbv2.describe_target_groups()   

        jsonResult = ""
        if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile: 
           jsonResult = Utils.dictToJson(target_groups)

        resultTxt = "" 
        targetGroups = target_groups['TargetGroups']
        if len(targetGroups) >= 1:
           resultTxt  = "Total of " + Style.GREEN + str(len(targetGroups)) + Style.RESET + " Target Groups\n"
           resultTxt  = resultTxt + Utils.separator()  + "\n\n"
           idx_lin    = 0
           header     = ["#","Target Group","VPC Id","Target Type","Protocol","Port","Load Balancer"]

           if healthInfo:
              header.append("HC Protocol")
              header.append("HC Enabled")
              header.append("HC Interval Sec")
              header.append("HC Path")
              header.append("HC Timeout Sec")
              header.append("Healthy Threshold Count")
              header.append("Unhealthy Threshold Count")
              header.append("Matcher")

           prettyTable    = PrettyTable(header)

           for targetGroup in targetGroups:
               idx_lin += 1

               labelTg = targetGroup['TargetGroupName']
               if showTgArn:
                  labelTg  = targetGroup['TargetGroupArn']

               labelElb = ""
               if len(targetGroup['LoadBalancerArns']) > 0:
                  if not showElbArn:
                     arnELB   = arnparse(targetGroup['LoadBalancerArns'][0])
                     labelElb = arnELB.resource
                  else:
                     labelElb = targetGroup['LoadBalancerArns'][0]


               columns = [ str(idx_lin), labelTg, \
                                         targetGroup['VpcId'], \
                                         targetGroup['TargetType'], \
                                         targetGroup['Protocol'], \
                                         targetGroup['Port'], \
                                         labelElb   
                         ]

               if healthInfo:
                  columns.append(targetGroup['HealthCheckProtocol'])            
                  columns.append(targetGroup['HealthCheckEnabled'])            
                  columns.append(targetGroup['HealthCheckIntervalSeconds'])
                  columns.append(targetGroup['HealthCheckPath'])
                  columns.append(targetGroup['HealthCheckTimeoutSeconds'])
                  columns.append(targetGroup['HealthyThresholdCount'])
                  columns.append(targetGroup['UnhealthyThresholdCount'])
                  if targetGroup['Matcher']['HttpCode']:
                     columns.append("HttpCode: " + targetGroup['Matcher']['HttpCode'])
                  else:   
                     columns.append("")

               prettyTable.addRow(columns)

           if (int(tableArgs.sortCol) - 1) > len(columns):
               tableArgs.sortCol = "1"
           prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
           prettyTable.ascendingOrder(not tableArgs.desc)
           result = ""
           if tgArn:
              if len(tgArn) > 1:
                 result  = result + "Target Groups: \n"
                 for tg in tgArn:
                     result  = result + "  - " + Style.GREEN + str(tg) + Style.RESET + "\n" 
                 result = result + "\n"
              else:   
                 result  = result + "Target Group...: " + Style.GREEN + str(tgArn) + Style.RESET + "\n\n"
           elif elbArn:
              result  = result + "Load Balancer..: " + Style.GREEN + elbArn + Style.RESET +  "\n\n"   

           if healthInfo:   
              result = result + "(HC) " + Style.GREEN + "Health Check" + Style.RESET + "\n\n"
           result = result + prettyTable.printMe(self.config.tableLineSeparator, tableArgs)
           return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_TARGET_GROUPS]["name"],result, self.config, jsonResult, False, tableArgs)
        else:
           return Utils.formatResult(Functions.FUNCTIONS[Functions.LIST_TARGET_GROUPS]["name"],resultTxt, self.config, "", False)     
    
    # A main function called from user
    # Find Load Balancers and Target Groups pointing to an EC2 Instance
    # arguments: instanceId
    def findInstancesElbsTgs(self):
        elbv2       = self.botoSession().client('elbv2')
        instanceId  = None
        showTgArn   = False
        showElbArn  = False
        tableArgs   = TableArgs()
        tableArgs.setArguments(self.config.commandArguments)
        
        if "," in self.config.commandArguments: 
           instanceId = self.config.commandArguments.split(",")[0]
           for arg in self.config.commandArguments.split(","):
               if "elbarn" in arg:
                  showElbArn = True
               elif "tgarn" in arg:
                  showTgArn = True      
        else:
           instanceId = tableArgs.cleanPipelineArguments()

        tableArgs.setArguments(self.config.commandArguments)

        if not instanceId:
           resultTxt = "Where is the Instance Id? You didn't tell me which EC2 Instance to look for... " + Emoticons.ops() +  Style.RESET
           return Utils.formatResult(Functions.FIND_ELBS_EC2,resultTxt, self.config, "", True, tableArgs)

        tableArgs.setArguments(self.config.commandArguments)

        target_groups = elbv2.describe_target_groups();

        dictResult = {}
        if len(target_groups) >= 1:
           resultTxt  = ""
           idx_lin    = 0 
           header = ["#","Instance Id","Load Balancer","Target Group", "Vpc Id", "Target Health"]

           prettyTable    = PrettyTable(header)
           printSeparator = False
           columns        = None

           for tg in target_groups['TargetGroups']:
                 targetsHealth = elbv2.describe_target_health(TargetGroupArn=tg["TargetGroupArn"])
                 for targetHealthDescription in targetsHealth["TargetHealthDescriptions"]:
                    if targetHealthDescription["Target"]["Id"] == instanceId:
                       idx_lin += 1

                       labelTg = tg['TargetGroupName']
                       if showTgArn:
                          labelTg  = tg['TargetGroupArn']
                       labelElb = ""
                       if len(tg['LoadBalancerArns']) > 0:
                          if not showElbArn:
                             arnELB   = arnparse(tg['LoadBalancerArns'][0])
                             labelElb = arnELB.resource
                          else:
                             labelElb = tg['LoadBalancerArns'][0]

                       labelTargetHealth = ""
                       if "State" in targetHealthDescription["TargetHealth"]:
                          if targetHealthDescription["TargetHealth"]["State"] == "healthy":
                             labelTargetHealth = "Healthy"
                          else:
                             labelTargetHealth = "Unhealthy"
                             #if "Reason" in targetHealthDescription["TargetHealth"]:
                             #    labelTargetHealth += ", " + targetHealthDescription["TargetHealth"]["Reason"]
                             if "Description" in targetHealthDescription["TargetHealth"]:
                                 labelTargetHealth += ", " + targetHealthDescription["TargetHealth"]["Description"]    

                       dictResult["instanceId"]   = instanceId
                       dictResult["loadBalancer"] = labelElb
                       dictResult["targetGroup"]  = labelTg
                       dictResult["vpcId"]        = tg["VpcId"]  
                       dictResult["targetHealth"] = labelTargetHealth         

                       columns = [ str(idx_lin), \
                                         instanceId, \
                                         labelElb, \
                                         labelTg, \
                                         tg["VpcId"], \
                                         labelTargetHealth
                                 ]
                       prettyTable.addRow(columns)                  

           jsonResult = ""
           if self.config.printResults or tableArgs.verbose or tableArgs.saveToFile:
              jsonResult = Utils.dictToJson(dictResult)            

           if columns:
               if (int(tableArgs.sortCol) - 1) > len(columns):
                     tableArgs.sortCol = "1"
               prettyTable.sortByColumn(int(tableArgs.sortCol) - 1)
               prettyTable.ascendingOrder(not tableArgs.desc)
               result = ""
               result = result + "EC2 Instance..: " + Style.GREEN + instanceId + Style.RESET +  "\n\n"   
               result = result + prettyTable.printMe(self.config.tableLineSeparator, tableArgs)
               return Utils.formatResult(Functions.FUNCTIONS[Functions.FIND_ELBS_EC2]["name"],result, self.config, jsonResult, False, tableArgs)
           else:
               resultTxt = "EC2 Instance..: " + Style.GREEN + instanceId + Style.MAGENTA + " wasn't found in any ELB/TargetGroup" + Style.RESET +  "\n"        
               return Utils.formatResult(Functions.FUNCTIONS[Functions.FIND_ELBS_EC2]["name"],resultTxt, self.config,"", False, tableArgs)
        else:
           resultTxt = Style.MAGENTA + " None Target Group was found! " + Style.RESET +  "\n"        
           return Utils.formatResult(Functions.FUNCTIONS[Functions.FIND_ELBS_EC2]["name"],resultTxt, self.config,"", False, tableArgs)
        
    def botoSession(self):
        return self.config.botoSession();

# For a Function Development
if __name__ == '__main__':
   config = Config()
   config.awsProfile = "ecomm"
   config.awsRegion = "eu-central-1"
   config.printResults = True
   #config.awsTags["Environment"] = "production"
   config.tableLineSeparator = False

   # subnets, is-public (only with subnets active)
   # sort1,desc (when requested together with subnets, the group lines will be omitted)
   #config.commandArguments = "health"
   #config.commandArguments = "sort3"
   #config.commandArguments = "arn:aws:elasticloadbalancing:eu-central-1:659915611011:loadbalancer/app/LB-Graylog/69198aad8d3f7668"
   #config.commandArguments = "arn:aws:elasticloadbalancing:eu-central-1:659915611011:loadbalancer/app/LB-Graylog/69198aad8d3f7668," + \
   #                          "arn:aws:elasticloadbalancing:eu-central-1:659915611011:targetgroup/TG-Graylog/3f4c285bea7970df"
      
   #config.commandArguments = "arn:aws:elasticloadbalancing:eu-central-1:659915611011:loadbalancer/app/LB-Graylog/69198aad8d3f7668," + \
   #                          "arn:aws:elasticloadbalancing:eu-central-1:659915611011:targetgroup/TG-Graylog/3f4c285bea7970df;" + \
   #                          "arn:aws:elasticloadbalancing:eu-central-1:659915611011:targetgroup/tf-20200602224957995200000001/0f8f9375912dfe3f,health"

   #config.commandArguments = "arn:aws:elasticloadbalancing:eu-central-1:659915611011:loadbalancer/app/LB-Graylog/69198aad8d3f7668,health,sort2"
   #config.commandArguments = "arn:aws:elasticloadbalancing:eu-central-1:659915611011:targetgroup/tf-20191023171348503100000005/d3cbc1b6a68a4fee"
   #config.commandArguments = "arn:aws:elasticloadbalancing:eu-central-1:659915611011:targetgroup/tf-20191023171348503100000005/d3cbc1b6a68a4fee;arn:aws:elasticloadbalancing:eu-central-1:659915611011:targetgroup/tf-20191023171348502800000004/ec3200de50acaed9,health"
   PyElbCP = PyElbCP(config)
   #report =  PyElbCP.findInstancesElbsTgs()
   report =  PyElbCP.listTargetGroups()
   Utils.addToClipboard(report)
   print(report)

   # To Do:
   # listLoadBalancers --> https://docs.aws.amazon.com/cli/latest/reference/elb/describe-load-balancers.html

   



import os, time
import logging
import sys, threading
import boto3
import configparser
from boto3.s3.transfer import TransferConfig
from botocore.exceptions import ClientError
from datetime import datetime
from os.path import expanduser
from tinydb import TinyDB, Query, where
from tinydb.storages import JSONStorage
from tinydb.middlewares import CachingMiddleware
import base64
from enum import Enum

LOG = logging.getLogger("app." + __name__)
LOG.setLevel(logging.INFO)

PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))
from PrettyTable import PrettyTable
from TableArgs import TableArgs

FILE_INI               = expanduser("~") + "/.pytransfers3/" + "pytransfers3.ini"
FILE_DB                = expanduser("~") + "/.pytransfers3/" + "#_pytransfers3.json"
SEPARATOR_FILES_UPLOAD = 112

# Check if the object exists
# aws s3api head-object --bucket www.codeengine.com --key index.html
#
# Dependencies
# pip install botocore
# pip install boto3
# pip install tinydb   or   grab it here "https://github.com/msiemens/tinydb/" and pip install .
#

class Style:
    UNDERLINE = '\033[4m'
    RESET     = '\033[0m'

    # Regular Colors
    BLACK     = '\033[30m'
    RED       = '\033[31m'
    GREEN     = '\033[32m'
    YELLOW    = '\033[33m'
    BLUE      = '\033[34m'
    MAGENTA   = '\033[35m'
    CYAN      = '\033[36m'
    WHITE     = '\033[37m'

    # Bold
    BBLACK    ="\033[30m"
    BRED      ="\033[31m"
    BGREEN    ="\033[32m"
    BYELLOW   ="\033[33m"
    BBLUE     ="\033[34m"
    BMAGENTA  ="\033[35m"
    BCYAN     ="\033[36m"
    BWHITE    ="\033[37m"

    # Background Colors
    BG_BLACK  = '\033[40m'
    BG_RED    = '\033[41m'
    BG_GREEN  = '\033[42m'
    BG_YELLOW = '\033[43m'
    BG_BLUE   = '\033[44m'
    BG_PURPLE = '\033[45m'
    BG_CYAN   = '\033[46m'
    BG_WHITE  = '\033[47m'

    # High Intensty
    IBLACK    ="\033[90m"
    IRED      ="\033[91m"
    IGREEN    ="\033[92m"
    IYELLOW   ="\033[93m"
    IBLUE     ="\033[94m"
    IMAGENTA  ="\033[95m"
    ICYAN     ="\033[96m"
    IWWHITE   ="\033[97m"

    # Bold High Intensty
    BIBLACK   ="\033[90m"
    BIRED     ="\033[91m"
    BIGREEN   ="\033[92m"
    BIYELLOW  ="\033[93m"
    BIBLUE    ="\033[94m"
    BIPURPLE  ="\033[95m"
    BICYAN    ="\033[96m"
    BIWHITE   ="\033[97m"

    # High Intensty backgrounds
    On_IBLACK ="\033[100m"
    On_IRED   ="\033[101m"
    On_IGREEN ="\033[102m"
    On_IYELLOW="\033[103m"
    On_IBLUE  ="\033[104m"
    On_IPURPLE="\033[;95m"
    On_ICYAN  ="\033[106m"
    On_IWHITE ="\033[107m"
    

class Emoticons:
    _PROMPT      = ['⛅','⛅']
    _WAITING     = ['💤💤💤','☕☕☕']
    _SEE_YA      = ['👋','✌']
    _ERROR       = ['❌💣','❌☠']
    _TOOL        = ['🔧','⚒']
    _THUMBS_UP   = ['👍','✔']
    _POINT_RIGHT = ['👉','❖']
    _WINK        = ['😉','☻']
    _OPS         = ['😕','☹']
    _PIN         = ['📌','✎']
    _ENV         = ['📝','✍']
    _TIME        = ['🕘','☕']
    _WAIT_DISTR  = ['🍺','♨']
    _WAIT_DISTR2 = ['🍼','⚾']
    _MAGNIFIER   = ['🔍','☌']
    _BLOCKS      = ['📦','❒']
    _REDMARK     = ['🔴','⚫']
    _UPLOAD      = ['📤','✈']
    _UPLOAD_PART = ['🔹','➩']
    _FOLDER      = ['🔹','➩']
    _OK          = ['✅','✅']
    _IMGS        = [
                     ['🕒','🕓','🕔','🕕','🕖','🕗','🕘','🕙','🕚','🕛','🕐','🕑'],
                     ['☰','☱','☲','☴','☵','☶','☷','☶','☴']
                   ]  
    @staticmethod
    def isWindows():
        return 1 if os.name == "nt" else 0
    
    @staticmethod
    def prompt():
        return Emoticons._PROMPT[Emoticons.isWindows()]
    @staticmethod
    def waiting():
        return Emoticons._WAITING[Emoticons.isWindows()]    
    @staticmethod
    def seeYa():
        return Emoticons._SEE_YA[Emoticons.isWindows()]
    @staticmethod
    def error():
        return Emoticons._ERROR[Emoticons.isWindows()]
    @staticmethod
    def tool():
        return Emoticons._TOOL[Emoticons.isWindows()]
    @staticmethod
    def thumbsUp():
        return Emoticons._THUMBS_UP[Emoticons.isWindows()] 
    @staticmethod
    def pointRight():
        return Emoticons._POINT_RIGHT[Emoticons.isWindows()]
    @staticmethod
    def wink():
        return Emoticons._WINK[Emoticons.isWindows()]
    @staticmethod
    def pin():
        return Emoticons._PIN[Emoticons.isWindows()]
    @staticmethod
    def env():
        return Emoticons._ENV[Emoticons.isWindows()]
    @staticmethod
    def time():
        return Emoticons._TIME[Emoticons.isWindows()]
    @staticmethod
    def waitDistract():
        return Emoticons._WAIT_DISTR[Emoticons.isWindows()]
    @staticmethod
    def waitDistract2():
        return Emoticons._WAIT_DISTR2[Emoticons.isWindows()]
    @staticmethod
    def ops():
        return Emoticons._OPS[Emoticons.isWindows()]
    @staticmethod
    def magnifier():
        return Emoticons._MAGNIFIER[Emoticons.isWindows()]
    @staticmethod
    def blocks():
        return Emoticons._BLOCKS[Emoticons.isWindows()]
    @staticmethod
    def redMark():
        return Emoticons._REDMARK[Emoticons.isWindows()]
    @staticmethod
    def upload():
        return Emoticons._UPLOAD[Emoticons.isWindows()]
    @staticmethod
    def uploadPart():
        return Emoticons._UPLOAD_PART[Emoticons.isWindows()]
    @staticmethod
    def folder():
        return Emoticons._FOLDER[Emoticons.isWindows()]                    
    @staticmethod
    def ok():
        return Emoticons._OK[Emoticons.isWindows()] 
    @staticmethod
    def imgs():
        return Emoticons._IMGS[Emoticons.isWindows()]                            

UploadState = Enum('UploadState', 'STARTED FINISHED')

class UtilTransferS3:
    @staticmethod
    def formatBytesView(bytes,rightAlign=True):
        result = "Ops..."
        if bytes >= 1000000000:
           bytesInMega = round(((bytes / 1024) / 1024) / 1024,2)
           result = format(bytesInMega,",") + "GB"
        elif bytes >= 1000000:
           bytesInMega = round((bytes / 1024) / 1024,2)
           result = format(bytesInMega,",") + "MB"
        elif bytes >= 1024:
           bytesInKbyte = round((bytes / 1024),2)
           result = format(bytesInKbyte,",") + "KB" 
        else:   
           result = format(bytes,",") + "b"  
        if rightAlign:   
           diff = 9 - len(result)   
           return  " ".ljust(diff if diff > 0 else 0," ") + result 
        return result   

class ProgressPercentage(object):

    def __init__(self, filename):
        self._filename = filename
        self._size     = float(os.path.getsize(filename))
        self._progress = 0
        self._lock     = threading.Lock()

    def __call__(self, bytes_amount):
        with self._lock:    
            self._progress += bytes_amount
            percentage = (self._progress / self._size) * 100

            msg = " ".ljust(42," ") +\
                  Style.IYELLOW + UtilTransferS3.formatBytesView(self._progress) + Style.GREEN + " of " + UtilTransferS3.formatBytesView(self._size) +\
                  "     (" + Style.IYELLOW + "{0:.2f}%".format(percentage) + Style.GREEN + ")"

            sys.stdout.write("\r     " + msg)
            sys.stdout.flush()

MULTIPART_UPLOAD      = "MULTIPART-UPLOAD"

class PyTransferS3:

    def __init__(self, useConfigAwsCockpit, configAwsCockPit, bucketName, profileName, regionName, prefix=None):
        self.bucketName                 = bucketName
        self.prefix                     = prefix
        self.target                     = ""
        self.multiPartUploadPreferences = None 
        self.database                   = None
        self.s3client                   = None
        self.s3Client                   = boto3.session.Session(profile_name=profileName,region_name=regionName).client("s3")
        self.partSizeBlockMegabytes     = 10
        ################################  Qt                            KB     MB
        self.partSizeBlockBytes         = self.partSizeBlockMegabytes * 1000 * 1000

        # Log init
        FILE_LOG = expanduser("~") + "/.pytransfers3/" + "log"
        if not os.path.exists(FILE_LOG):
           os.makedirs(FILE_LOG)

        rootLogger = logging.getLogger()
        logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")

        fileHandler = logging.FileHandler("{0}/{1}.log".format(FILE_LOG, "py-transfers3"))
        fileHandler.setFormatter(logFormatter)
        fileHandler.setLevel(logging.WARN)
        rootLogger.addHandler(fileHandler)

        # Config init
        if not useConfigAwsCockpit:
           if os.path.exists(FILE_INI):
              config = configparser.ConfigParser(allow_no_value=True)
              config.read(FILE_INI)
              self.multiPartUploadPreferences = config[MULTIPART_UPLOAD]["threshold"]
           else:   
              config = configparser.ConfigParser(allow_no_value=True)
              config.add_section(MULTIPART_UPLOAD)
              config.set(MULTIPART_UPLOAD, "# The size is in Megabytes")
              config.set(MULTIPART_UPLOAD, "threshold","100")
              with open(FILE_INI,'w') as configfile:
                  config.write(configfile)
              config = configparser.ConfigParser(allow_no_value=True)
              config.read(FILE_INI)
              self.multiPartUploadPreferences = config[MULTIPART_UPLOAD]["threshold"]
        else:
            self.partSizeBlockMegabytes     = int(configAwsCockPit.uploadChunkSizeMultipart)
            #####################################################  Qt            KB     MB
            self.partSizeBlockBytes         = int(self.partSizeBlockMegabytes) * 1000 * 1000
            self.multiPartUploadPreferences = configAwsCockPit.uploadThresholdMultipart

    def initDb(self):
        # DB init
        databaseFileName = FILE_DB.replace("#",self.encodeString(self.target))
        self.database = TinyDB(databaseFileName, indent=4)

        # Use this in case a high intesive I/O operations (to cache a little), but with caution!
        # you should use the method below to garantee that the operations are recorded and not lost consistency:   
        #  with self.database as db:  
        #       db.insert(...)
        #self.database = TinyDB(databaseFileName, indent=4, storage=CachingMiddleware(JSONStorage))   

    def encodeString(self, str):
        bytes        = str.encode("utf-8")
        base64_bytes =  base64.b64encode(bytes)
        return base64_bytes.decode("utf-8")
    def decodeString(self, str):
        bytes        = str.encode("utf-8")
        base64_bytes =  base64.b64decode(bytes)
        return base64_bytes.decode("utf-8")

    def transfer(self, recursive):
        if not self._checkBucketExist(self.bucketName):
           msg = "\n" + Style.IRED + Emoticons.error() + Style.GREEN + " The Bucket \033[95m{}\033[32m does not exit...".format(self.bucketName) + "\n"
           print(msg)
           return ""

        if self.target[1:2] == ":":
           # At Windows convert driver letter to Uppercase (it fails in Lowercase)
           self.target = self.target[0:1].upper() + ":" + self.target[2:] 
        elif self.target[0:1] == "~":
           # At Linux, expand the user home alias
           self.target = expanduser("~") + self.target[1:]

        self.initDb()
        isOnlyAFile = self.isFile(self.target)
        print("")

        if isOnlyAFile:
            # Transfer file
            file     = self.target[self.target.rfind("/")+1:]
            filePath = self.target.replace(file,"")
            keyPath  = file
            self.checkUploadFile(keyPath, self.target)
        else:
            if recursive:    
               # Transfer everything inside each folder
                for root, dirs, files in os.walk(self.target):
                    for file in files:
                        filePath = os.path.join(root, file)
                        keyPath  = self.defineKeyPathFromTarget(root,file)
                        self.checkUploadFile(keyPath, filePath)   
            else:
               # Transfer only the files inside the requested folder
                for file in os.listdir(self.target):
                    if self.isFile(self.target + "/" + file):
                        # Transfer file 
                        keyPath  = self.defineKeyPathFromTarget(self.target,file)
                        self.checkUploadFile(keyPath, self.target + "/" + file)

        self.database.close()
        print("")                

    def checkUploadFile(self, keyPath, filePath):
        Upload = Query()
        upload = self.database.get(Upload.key == keyPath)

        if upload and "finish" in upload and upload["finish"] != "":
           self.showMessage(Emoticons.pin() + " File " + Style.BLUE + filePath + Style.GREEN + \
                        " was already transfered to " + Style.BLUE + "s3://" + self.bucketName + "/" + keyPath + Style.GREEN +\
                        " at " + Style.BLUE + upload["finish"] + "\n         \033[33mSize " + "\033[34m~{0:.1f}MB\033[0m".format(upload['file']['length'] / 1024 / 1024))
           msg = " \033[34m--->\033[33m " + Emoticons.pin() + "\033[32m Maybe it was delete from S3 Bucket, and we did not know yet, so... \n         Would like to \033[94mgo ahead\033[33m anyway ? " + \
                 "[\033[94mY\033[33m/\033[94mN\033[33m]: "                
           val = input(msg)       
           if val.upper() == "Y":
              # Clean DB's record of the Upload
              self.database.remove(where('key') == keyPath) 
              upload = self.database.get(Upload.key == keyPath)
              print(" \033[34m--->\033[0m " + Emoticons.pin() + " Ok, let's do it...\n")
              self.transferFile(keyPath,filePath,upload)
        else:
           self.transferFile(keyPath,filePath,upload)

    def removeTargetHistory(self):
        target = self.encodeString(self.target) + "_pytransfers3.json"
        dir    = expanduser("~") + "/.pytransfers3/"
        # Clean DB
        records = self.database.all()
        for r in records:
            self.database.remove(where('key') == r['key']) 
        self.database.close()    
        filePath = dir + target
        os.remove(filePath)
        print("")
        self.showMessage(" Target history for \033[94m{}\033[33m deleted!".format(self.target))
        print("")

    def showTargets(self):
        tableArgs   = TableArgs()
        dir         = expanduser("~") + "/.pytransfers3/"
        header      = ["#","Target"]
        prettyTable = PrettyTable(header)
        idx_lin     = 1

        print("")
        self.showMessage(" Targets you have used for Uploads")
        for root, dirs, files in os.walk(dir):
            files = [ fi for fi in files if fi.endswith(".json") ]
            for file in files:
                target = self.decodeString(file.replace("_pytransfers3","").replace(".json",""))
                columns    = [ str(idx_lin), target ]
                prettyTable.addRow(columns)
                idx_lin = idx_lin + 1
        prettyTable.sortByColumn(0)   
        prettyTable.numberSeparator = True
        prettyTable.ascendingOrder(True)
        print(prettyTable.printMe(False, tableArgs))
        self.showMessage(" You can see more info using \033[95mshowTargetUploads \033[94m[TARGET]\033[0m")
        self.showMessage(" You can delete the target history \033[95mremoveTargetHistory \033[94m[TARGET]\033[0m")
        print("")

    def showTargetUploads(self, showAllData=False):
        tableArgs  = TableArgs()
        records = self.database.all()

        multiPartListed = False
        keySample       = None

        print("")
        self.showMessage(" Upload registry for target \033[94m{}\033[33m".format(self.target))
        if len(records) > 0:
           if showAllData:
              header       = ["#","Key", "Bucket","Start","Finish","State","UploadId","Part Size","Total Parts","File Name","File Length"]
           else:
              header       = ["#","Key", "Bucket","Start","Finish","Parts"]
           prettyTable  = PrettyTable(header)
           idx_lin      = 1
           for r in records:
               multiParts = len(r["parts"])
               if multiParts > 0:
                  keySample = r["key"]
                  multiPartListed = True
               if showAllData:
                  columns    = [ str(idx_lin), r["key"], r["bucket"], r["start"], r["finish"], r["state"], r["uploadId"], 
                                 r["partSize"],multiParts,r["file"]["name"], r["file"]["length"] ]
               else:    
                  if multiParts == 0:
                     multiParts = 1
                  columns    = [ str(idx_lin), r["key"], r["bucket"], r["start"], r["finish"], multiParts ]
   
               prettyTable.addRow(columns)
               idx_lin = idx_lin + 1
   
           prettyTable.sortByColumn(0)   
           prettyTable.numberSeparator = True
           prettyTable.ascendingOrder(True)
           print(prettyTable.printMe(False, tableArgs))
           if multiPartListed:
              self.showMessage(" You can view parts of a MultiPart Upload using \n       \033[95mshowTargetUploadParts \033[94m{}\033[33m,\033[94m{}\033[0m".format(self.target,keySample))
              print("")
        else:
           self.showMessage(" No records found!")

    def showTargetUploadsParts(self, key):
        tableArgs  = TableArgs()
        records = self.database.search(where('key') == key)

        print("")
        self.showMessage(" Uploaded Parts registry for target \033[94m{}\033[33m key:\033[94m{}\033[33m".format(self.target,key))
        if len(records) > 0 and len(records[0]["parts"]) > 0:
           self.showMessage(" Total of parts uploaded \033[94m{}\033[33m".format(len(records[0]["parts"]))) 
           header       = ["#","PartNumber", "ETag","Uploaded Bytes"]
           prettyTable  = PrettyTable(header)
           idx_lin      = 1
           for r in records:
               for p in r["parts"]:
                   columns = [ str(idx_lin), p["PartNumber"], p["ETag"], p["UploadedBytes"] ]
                   prettyTable.addRow(columns)
                   idx_lin = idx_lin + 1
   
           prettyTable.sortByColumn(0)   
           prettyTable.numberSeparator = True
           prettyTable.ascendingOrder(True)
           print(prettyTable.printMe(False, tableArgs))
        else:
           self.showMessage(" No parts records found for \033[94m{}\033[0m".format(key))       


    def createUpload(self, keyPath, file, length, uploadId):
        partSize = self.partSizeBlockBytes
        # In case it is not a MultipartUpload, this value will not exist
        if not uploadId:
           uploadId  = ''
           partSize  = 1

        self.database.insert({
            'key': keyPath,
            'bucket':self.bucketName,
            'start': self.dateTimeNow(),
            'finish': '',
            'state': UploadState.STARTED.value,
            'uploadId':uploadId,
            'partSize': partSize,
            'parts':[],
            'file':{
                'name': file,
                'length': length,
            }
        })

    def updateMultiPartUpload(self, keyPath, part):
        Upload = Query()
        upload = self.database.get(Upload.key == keyPath)
        if not upload:
           raise ValueError(Emoticons.error() + ' The already start MultiPartUpload with the KeyPath=' + keyPath + ' was not found! It should be there!!')
        upload['parts'].append(part)
        self.database.update(upload, Upload.key == keyPath)

    def finishUpload(self, keyPath):
        Upload = Query()
        upload = self.database.get(Upload.key == keyPath)
        if not upload:
           raise ValueError(Emoticons.error() + '  The already Upload with the KeyPath=' + keyPath + ' was not found! It should be there!!')
        upload['finish'] = self.dateTimeNow()
        upload['state'] = UploadState.FINISHED.value
        self.database.update(upload, Upload.key == keyPath)

    def cleanNotFinishedUpload(self, keyPath):
        self.database.remove(where('key') == keyPath)    

    def transferFile2(self,keyPath,file):
        print("----")    

    def _checkBucketExist(self,bucketName):
        try:   
            self.s3Client.list_objects_v2(Bucket=bucketName, Delimiter = "/")
            return True
        except ClientError as ex:
            return False

    def showHeaderInfoUplad(self, isMultipart, keyPath, file, upload):
        print("-".ljust(SEPARATOR_FILES_UPLOAD,"-"))
        msg = Emoticons.tool() + " UPLOAD SETUP" + "\n" + \
              "      " + Emoticons.pointRight() + " Chunk Size....................: \033[34m {}MB\033[32m".format(self.partSizeBlockMegabytes) + "\n" + \
              "      " + Emoticons.pointRight() + " Threashold Multipart Uploads..: \033[34m{}MB\033[32m".format(self.multiPartUploadPreferences) + "\n\n" + \
              "      " + "   Hit \033[94mCTRL+C\033[32m to \033[94mcancel/pause\033[32m the download, you can \033[34mresume\033[32m later, executing the same command again"   
        self.showMessage(msg)
        print("-".ljust(SEPARATOR_FILES_UPLOAD,"-"))
        print("      " + Style.GREEN + Emoticons.pointRight() + " Uploading... \033[94m{}\033[33m to \033[94ms3://{}\033[33m".format(file.replace("\/","\\"),self.bucketName) )
        print("")

    def transferFile(self,keyPath,file,upload):
        try:
            if os.path.exists(file):
               totalBytes             = os.stat(file).st_size
               fileSizeAboveThreshold = True if ((totalBytes / 1024) / 1024) >= int(self.multiPartUploadPreferences) else False
   
               # Multipart upload already start, must continue where it stops
               # File size bigger than the threshold, starting a multipart upload
               if (upload and "uploadId" in upload) or fileSizeAboveThreshold:
                   self.showHeaderInfoUplad(True,keyPath,file,upload)
                   self.transferFileUsingMultipartUpload(keyPath, file, totalBytes, upload)
               else:
                   self.showHeaderInfoUplad(False,keyPath,file,upload)
                   self.transferFileAtOnce(keyPath, file, totalBytes)
            else:
               self.showMessage(Emoticons.error() + " Path \033[91m" + file + "\033[32m was not found!") 

        except:
            LOG.error("Unexpected error:" + str(sys.exc_info()[0]))
            self.showMessage(Emoticons.error() + " Unexpected error:\033[91m" + str(sys.exc_info()[0]) + "\033[32m")
            raise

    def calculateQtdeParts(self,totalBytes,partSize):    
        qtdeParts = int(totalBytes / partSize)
        if ((totalBytes / partSize) - int(totalBytes / partSize)) > 0:
            qtdeParts += 1
        return qtdeParts   

    def _messageAndRemoveNotFinishUploadWithUploadId(self,uploadId):
        print("") 
        self.showMessage(Emoticons.error() + " Ops... " + Style.RED + "Error! " + Style.GREEN + "Sorry, MultiPartUpload with id\n      " + Style.BLUE + uploadId + Style.GREEN + \
               " wasn't found!\n      Probably someone has deleted the unfinished Multipart Upload registry from AWS S3, unfortunately you'll have to start it again from the beginning." + \
               "\n      This entry was delete from your database, please relaunch the command and your Upload will start.")
        self.database.remove(where('uploadId') == uploadId)
        print("")

    def transferFileUsingMultipartUpload(self,keyPath,file,totalBytes,upload):        
        totalBytesUploaded   = 0
        totalBytesRemaning   = 0
        partSequenceUploaded = 0
        partSize             = self.partSizeBlockBytes
        qtdeParts            = 0

        resumingUpload = True if upload else False

        if resumingUpload:
           partSize             = upload["partSize"]
           partAlreadyUploaded  = upload["parts"]
           totalBytesUploaded   = 0
           qtdeParts            = self.calculateQtdeParts(totalBytes,partSize)
           partSequenceUploaded = len(partAlreadyUploaded)
           for p in partAlreadyUploaded:
               totalBytesUploaded    += int(p["UploadedBytes"])
           totalBytesRemaning = totalBytes - totalBytesUploaded
           self.message(Emoticons.upload() + " Resuming uploading in " + format(qtdeParts,",") + " blocks " + Emoticons.blocks() + " of " + UtilTransferS3.formatBytesView(partSize,False) + \
                        ", remaining " + Style.BLUE + UtilTransferS3.formatBytesView(totalBytesRemaning,False) + Style.GREEN + " of " +\
                        Style.BLUE + UtilTransferS3.formatBytesView(totalBytes,False) + Style.GREEN + ", file\n      " + Style.IBLUE + file + Style.GREEN)
        else:
           qtdeParts = self.calculateQtdeParts(totalBytes,partSize)    
           self.message(Emoticons.upload() + " Uploading in " + format(qtdeParts,",") + " blocks " + Emoticons.blocks() + " of " + UtilTransferS3.formatBytesView(partSize,False) + \
                        " a total of " + Style.BLUE + UtilTransferS3.formatBytesView(totalBytes,False) + Style.GREEN + ", file \n" \
                        + Style.IBLUE + "      " + file + Style.RESET)

        doItUpload  = False
        mpu         = None
        mpu_id      = None
        ## create upload registry
        if not resumingUpload:
           # Create MPU
           mpu    = self.s3Client.create_multipart_upload(Bucket=self.bucketName, Key=keyPath)
           mpu_id = mpu["UploadId"]
           
           self.createUpload(keyPath, file, totalBytes, mpu_id)

           doItUpload = True
        else:
           #mpu_id = upload["uploadId"] 
           #doItUpload = True

           # Here get Upload (check if exists)   
           mpus = self.s3Client.list_multipart_uploads(Bucket=self.bucketName)
           if "Uploads" in mpus:
               for u in mpus["Uploads"]:
                   if u["UploadId"] == upload["uploadId"]:
                      mpu_id = upload["uploadId"]
                      doItUpload = True   
                      break
               if not doItUpload:
                  self._messageAndRemoveNotFinishUploadWithUploadId(upload["uploadId"])
           else:
               self._messageAndRemoveNotFinishUploadWithUploadId(upload["uploadId"])

        if doItUpload:
           # Upload Parts
           parts = []
           if resumingUpload:
              # Add Parts already upload before 
              for part in upload["parts"]:
                  parts.append({"PartNumber": part["PartNumber"], "ETag": part["ETag"]})
           uploadedBytes = totalBytesUploaded
           imgIdx = 0
           imgs   = Emoticons.imgs()
           #imgs   = ['🕒','🕓','🕔','🕕','🕖','🕗','🕘','🕙','🕚','🕛','🕐','🕑']
           with open(file, "rb") as f:
               # Position at the part we still have to upload (in case first time, the value will be zeros)
               f.seek(uploadedBytes) 
               i = partSequenceUploaded + 1

               firstLoop = True
               while True:
                   data = f.read(partSize)
                   if not len(data):
                      break
   
                   sys.stdout.write("\r      " + Emoticons.uploadPart() + " " + Style.GREEN + "Uploading part... " + Style.IYELLOW + format(i,",") + Style.GREEN + " of " \
                                    + Style.IYELLOW + format(qtdeParts,",") + Style.RESET)
                   if firstLoop:
                      #time.sleep(1) 
                      firstLoop = False                  
   
                   part = self.s3Client.upload_part(Body=data,
                                               Bucket=self.bucketName,
                                               Key=keyPath,
                                               UploadId=mpu_id,
                                               PartNumber=i)
   
                   parts.append({"PartNumber": i, "ETag": part["ETag"]})
                   # Update Upload Part
                   self.updateMultiPartUpload(keyPath,{
                       "PartNumber": i, 
                       "ETag": part["ETag"],
                       "UploadedBytes": len(data)
                   })
                   uploadedBytes += len(data)
                   msg = Style.IYELLOW + UtilTransferS3.formatBytesView(uploadedBytes) + Style.GREEN + " of " + Style.GREEN + UtilTransferS3.formatBytesView(totalBytes)
                   msg = msg + Style.GREEN + "     " + \
                         "(" + Style.IYELLOW  + "{0:.2f}%".format(self.asPercent(uploadedBytes, totalBytes)) + Style.GREEN + ")" + Style.RESET
                   sys.stdout.write("\r                                     " + imgs[imgIdx] + "        " + msg)                                              
                   #print(msg)
                   i += 1
                   imgIdx +=1
                   if (imgIdx+1) > len(imgs):
                      imgIdx = 0

               sys.stdout.write("\r      " + Emoticons.ok() + " " + Style.GREEN + "Finished " + Style.RESET) 
               print(Style.RESET + "\n-".ljust(SEPARATOR_FILES_UPLOAD,"-"))      
   
           # Complete MPU
           result = self.s3Client.complete_multipart_upload(Bucket=self.bucketName, Key=keyPath,\
                                                       UploadId=mpu_id,
                                                       MultipartUpload={"Parts": parts})
           self.finishUpload(keyPath)

           return result

    def dateTimeNow(self):
        return datetime.now().strftime("%Y-%m-%d %H:%M:%S:%f")

    def defineKeyPathFromTarget(self,dir,file):
        keyPath = dir.replace(self.target,"") + "/" + file
        return self.keyPathCleanFirstSlash(keyPath)

    # Helper
    def asPercent(self,num, denom):
        return float(num) / float(denom) * 100.0

    def transferFile3(self,keyPath,file):
        print(self.bucketName  + keyPath)

    def keyPathCleanFirstSlash(self, keyPath):
        if keyPath.startswith("/"):
           return keyPath[1:]
        return keyPath   

    def transferFileAtOnce(self,keyPath,file,totalBytes):     
        # Simple Upload
        config = TransferConfig(
                   multipart_threshold=1024 * 25,\
                   max_concurrency=10,\
                   multipart_chunksize=1024 * 25,\
                   use_threads=True)
        try:
            self.createUpload(keyPath, file, totalBytes, None)
            self.message(Emoticons.upload() + " Uploading file \n      " + Style.IBLUE + file + Style.GREEN)
            self.s3Client.upload_file(file, self.bucketName, keyPath,
                               #ExtraArgs={'ACL':'public-read','ContentType': 'text/pdf'},
                               Config=config,
                               Callback=ProgressPercentage(file))   
            self.finishUpload(keyPath)
            print(Emoticons.ok() + " " + Style.GREEN + "Finished " + Style.RESET) 
        except:
            self.cleanNotFinishedUpload(keyPath)
            print(" " + Emoticons.error() + " " + Style.RED + "Error " + Style.RESET + "{0} when uploadinf the file {1}...".format(sys.exc_info()[0],file))
            raise
        print(Style.RESET + "-".ljust(SEPARATOR_FILES_UPLOAD,"-"))      
        print("")

    def abortUploads(self,printMe=True):
        result            = {}
        found             = True
        result["bucket"]  = self.bucketName
        result["found"]   = True
        try:
            mpus   = self.s3Client.list_multipart_uploads(Bucket=self.bucketName)
        except ClientError as ex:
            #reason = ex.response['Error']['Code']
            found = False
            result["found"] = False

        if not found:
            result["found"] = False
            if printMe:
               print("")
               self.message(Emoticons.pin() + " Bucket not found!")        
               print("")
        elif "Uploads" in mpus:
            total = len(mpus["Uploads"])   

            result["nothing"] = False
            result["total"]   = total
            result["uploadsAborted"] = []

            if printMe:
               print("") 
               self.message(Emoticons.pin() + " ABORTING... " + str(total) +  (" UPLOADS:" if total > 1 else " UPLOAD:"))
            id = 0
            for u in mpus["Uploads"]:
                id += 1
                uploadId  = u["UploadId"]
                key       = u["Key"]
                msgResult = self.s3Client.abort_multipart_upload(Bucket=self.bucketName, Key=key, UploadId=uploadId)

                result["uploadsAborted"].append({
                    "id":str(id),
                    "uploadId":u["UploadId"],
                    "key":u["Key"]
                })

                if printMe:
                   self.message("   " + Emoticons.redMark() + " [" + str(id) + "] Aborted " + Style.BLUE + u["UploadId"] + "-" + u["Key"] + Style.GREEN)
            if printMe:       
               print("")        
        else:
            if printMe:
               print("")
               self.message(Emoticons.pin() + " Nothing to abort")        
               print("")
            result["nothing"] = True
        return result    

    def listUploads(self):    
        result = ""
        id     = 0
        mpus   = None
        found  = True
        if self.prefix and self.prefix != "":
           self.bucketName = self.bucketName.replace("/","") 
           self.prefix = "" if self.prefix == "/" else self.prefix
           if self.prefix.startswith("/"):
              self.prefix = self.prefix[1:]
           mpus = self.s3Client.list_multipart_uploads(Bucket=self.bucketName,Delimiter="/",Prefix=self.prefix)   
        else:      
           try: 
               mpus = self.s3Client.list_multipart_uploads(Bucket=self.bucketName)
           except ClientError as ex:
               #reason = ex.response['Error']['Code']
               found = False

        if not found:    
            result += "\n"
            result += self.message(Emoticons.pin() + " The bucket \033[35m{}\033[32m was not found".format(self.bucketName))
            result += "\n"
        elif "Uploads" in mpus:
            result += "\n"
            total = len(mpus["Uploads"])   
            result += self.message(Emoticons.pin() + " LISTING total of " + str(total) +  (" UPLOADS:" if total > 1 else " UPLOAD:"))
            for u in mpus["Uploads"]:
                id += 1
                result += self.message(Emoticons.folder() + " [" + "{:02d}".format(id) + "] " + Style.BLUE + u["Key"] + Style.GREEN + " - " + u["Initiated"].strftime("%Y-%m-%d %H:%M") + " - " + u["UploadId"])
            result += "\n"
        else:
            result += "\n"
            result += self.message(Emoticons.pin() + " There's no Multipart Upload in course")
            result += "\n"

        return {
            "foundBucket": found,
            "jsonResult":mpus,
            "result":result,
            "total": id
        }    

    def isFile(self, p):
        return not os.path.isdir(p)

    def message(self, msg, level=0):
        #print(Style.BLUE + " ---> " + Style.GREEN + msg + Style.RESET)    
        return Style.BLUE + " ---> " + Style.GREEN + msg + Style.RESET + "\n"

    def showMessage(self, msg, level=0):
        print(Style.BLUE + " ---> " + Style.GREEN + msg + Style.RESET)    


# python pytransfers3.py transfer,uploadsujr,/home/osboxes/Developer/tmp-upload/,recursive
# python pytransfers3.py transfer,uploadsujr,/home/osboxes/Developer/tmp-upload/dir1/,recursive
# python pytransfers3.py transfer,uploadsujr,/home/osboxes/Developer/tmp-upload/myfile2.tar
# python pytransfers3.py transfer,uploadsujr,/home/osboxes/Developer/tmp-upload/myfile2.tar
# python pytransfers3.py transfer,uploadsujr,/home/osboxes/Developer/tmp-upload/dir2
# python pytransfers3.py listuploads
# python pytransfers3.py abortuploads

def syntax():
    print("")
    print(Style.GREEN + "--------------------------------------------------------------------------------------------------------------------------")
    print(Style.GREEN + "  Available Commands: " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "[listuploads|abortuploads|transfer|showTargets|showTargetUploads|showTargetUploadsParts|removeTargetHistory]")
    print(Style.GREEN + "  Usages: ")
    print(Style.GREEN + "         " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "listuploads BUCKET_NAME")     
    print(Style.GREEN + "         " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "abortuploads BUCKET_NAME")
    print(Style.GREEN + "         " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "transfer BUCKET_NAME,FILE|FOLDER,[recursive]")     
    print(Style.GREEN + "         " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "showTargets")
    print(Style.GREEN + "         " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "showTargetUploads FILE|FOLDER")
    print(Style.GREEN + "         " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "showTargetUploads FILE|FOLDER,showAll")
    print(Style.GREEN + "         " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "showTargetUploadsParts FILE|FOLDER,KEY")     
    print(Style.GREEN + "         " + Style.WHITE + "python " + Style.GREEN + "pytransfer.py " + Style.BLUE + "removeTargetHistory TARGET")
    print(Style.GREEN + "--------------------------------------------------------------------------------------------------------------------------")
    print(Style.RESET)

if __name__ == '__main__':
    if "transfer" in sys.argv[1] and len(sys.argv[1].split(",")) <= 2:
       syntax()
       sys.exit()
    
    command = sys.argv[1].split(",")[0].lower()
    bucket  = ""
    
    pyTransferS3 = PyTransferS3(False,None,bucket,"ecomm","eu-central-1",None)
    if "listuploads" in command:
        pyTransferS3.bucketName = sys.argv[1].split(",")[1]
        pyTransferS3.listUploads()
    elif "abortuploads" in command:
        pyTransferS3.bucketName = sys.argv[1].split(",")[1]
        pyTransferS3.abortUploads()    
    elif "transfer" in command:
        pyTransferS3.bucketName = sys.argv[1].split(",")[1]
        path      = sys.argv[1].split(",")[2]
        recursive = True if "recursive" in sys.argv[1] else False
        pyTransferS3.bucketName = bucket
        pyTransferS3.target     = path
        pyTransferS3.transfer(recursive)
    elif "showTargets".lower() == command.lower():
        pyTransferS3.showTargets()
    elif "showTargetUploads".lower() == command.lower():
        if "," in sys.argv[2]:
           target      = sys.argv[2].split(",")[0]
           showAllData = True if sys.argv[2].split(",")[1].lower() == "showAll".lower() else False
        else:
           target      = sys.argv[2]
           showAllData = False
        pyTransferS3.target = target #"D:\\Downloads\\test\\" "d:/Temp/file.zip"
        pyTransferS3.initDb()
        pyTransferS3.showTargetUploads(showAllData)
    elif "showTargetUploadsParts".lower() == command.lower():
        target    = sys.argv[2].split(",")[0]
        key       = sys.argv[2].split(",")[1]
        pyTransferS3.target = target #"D:\\Downloads\\test\\" "d:/Temp/file.zip"
        pyTransferS3.initDb()
        pyTransferS3.showTargetUploadsParts(key)
    elif "removeTargetHistory".lower() == command.lower():
        target    = sys.argv[2].split(",")[0]
        pyTransferS3.target = target #"D:\\Downloads\\test\\" "d:/Temp/file.zip"
        pyTransferS3.initDb()
        pyTransferS3.removeTargetHistory()
    else:
        print("Command " + Style.GREEN  + command + Style.RESET + " not recognized")    
  
